      SUBROUTINE KarWdec(mode,Ninp,Nout,IdGen,aml,amn,QW,Ecm,Emin,q1,q2,
     &                   thel,phil,ql,qn,nphot,photk,phsum,ys,zs,Wt)
!     *****************************************************************
!----------------------------------------------------------------------!
! YFS-type MC generator of multiphoton radiation leptonic in W-decay:  !
!                 W(Q) --> l(ql) + nu(qn) + n*gamma(k_i)               !
! where for W-: l - charged lepton, nu - anti-neutino;                 !
!       for W+: l - charged anti-lepton, nu - neutino;                 !
! NOTE: Generation is done in the W rest frame (WRF) and the resulting !
!       4-momenta are given in that frame!                             !
! INPUT: mode =-1/0/1 -- initialization/generation/post-generation     !
!        Ninp,Nout -- I/O unit numbers (for printouts)                 !
!        IdGen -- generator ID for histograms and weight monitoring    !
!        aml,amn -- masses of final-state charged lepton and neutrino  !
!        QW = -1/+1 for W-/W+ (electric charge of W in units od e)     !
!        Ecm -- CMS energy of hadronic collision                       !
!        Emin -- soft photon cut-off, i.e. E_gamma^min [GeV]           !    
!        q1,q2 -- 4-momenta of quark and antiquark, resp.              !
!        thel,phil -- polar and azimuthal angles of lepton in WRF frame!
! OUPUT: ql,qn -- 4-momenta of final-state charged lepton and neutrino !
!        nphot,photk,phsum -- photons multiplicity, 4-momenta,         !
!                             sum of 4-momenta, resp.                  !
!        ys,zs - Sudakov variables                                     !
!        Wt -- weight from radiative W-decay generation                !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek                         CERN, July 2002  !
! Last update: 23.01.2015             by: W.P. (Krakow)                !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      PARAMETER( alfpi=  1/pi/alfinv, alfa= 1d0/alfinv)
      PARAMETER( NphMax =  100 )
      DIMENSION q1(4),q2(4)
      DIMENSION ql(4),qn(4),phsum(4),photk(NphMax,4)
      DIMENSION ys(NphMax),zs(NphMax)
! W 4-momentum (in WRF)
      DIMENSION Q(4),Qi(4)
      DIMENSION Xmfe(20),Xckm(3,3),pqu(4),paq(4),qd(4),qu(4),qk(4)
      INTEGER Ivec(6)    
* BoX-FORMATs for nice and flexible outputs
      CHARACTER*80 bxope,bxclo,bxtxt,bxl1i,bxl1l,bxl1f,bxl2f,bxl1g,bxl2g
      DATA NWtYFS /0/
      SAVE
!=======================================================================
!======================= Initialization mode ===========================
!=======================================================================
      IF (mode.EQ.-1) THEN
         XevGen = 0
! Print-out level
        CALL WH_GetLevPrn(LevPrn)
*/ Get quark masses
        CALL WH_GetmfCKM(Xmfe,Xckm)
! Infrared cut in CMS
        epsk = 2*Emin/Ecm
! Multiphoton radiation if MltFSR=0 - normal setting.
        MltFSR = 0
! Internal switch (inactive!)
        KeyRem = 0
!
*/ Get box-formats
         CALL WH_GetBxFmts(
     &          bxope,bxclo,bxtxt,bxl1i,bxl1l,bxl1f,bxl2f,bxl1g,bxl2g)
        WRITE(Nout,bxope)
        WRITE(Nout,bxtxt) 'Initialize KarWdec - START'
        IF (QW.LT.0) THEN
           WRITE(Nout,bxtxt) ' ===> W- decay '
        ELSE
           WRITE(Nout,bxtxt) ' ===> W+ decay '
        ENDIF
        CALL gbook1(IdGen+20,'KarWdec: log10(k/eps) marked$',50,0d0,5d0)
! Weight monitoring
        CALL gmonit(-1,IdGen+60,0d0,1d0,1d0)
        CALL gmonit(-1,IdGen+61,0d0,1d0,1d0)
        CALL gmonit(-1,IdGen+62,0d0,1d0,1d0)
        CALL gmonit(-1,IdGen+63,0d0,1d0,1d0)
        CALL gmonit(-1,IdGen+64,0d0,1d0,1d0)
!WP        CALL gmonit(-1,IdGen+65,0d0,1d0,1d0)
!WP        CALL gmonit(-1,IdGen+66,0d0,1d0,1d0)
        CALL gmonit(-1,IdGen+69,0d0,1d0,1d0)
        WRITE(Nout,bxtxt) 'Initialize KarWdec - END  '
        WRITE(Nout,bxclo)
!WP: Monitoring of QED interference weight
!WP        CALL gbook1(IdGen+110,'KarWdec: Interference wgt$',50,-2d0,3d0)


!=======================================================================
!======================== Generation mode ==============================
!=======================================================================
      ELSEIF (mode.EQ.0) THEN
        XevGen = XevGen + 1
        Wt = 0d0
! W invariant mass and 4-momentum in WRF
        aM2 = ( q1(4) + q2(4) )**2 - ( q1(3) + q2(3) )**2
     &      - ( q1(2) + q2(2) )**2 - ( q1(1) + q2(1) )**2
        aM   = SQRT(aM2)
        Q(4) = aM
        Q(3) = 0d0
        Q(2) = 0d0
        Q(1) = 0d0
! Soft photon cut-off for real photon generation at low level
! (i.e. in W rest frame)
        Ebeam = 0.5*Ecm
        x1  = q1(4)/Ebeam
        x2  = q2(4)/Ebeam
        xx = MAX(x1,x2)
        delta = epsk/xx
        EminWrf = delta * aM/2
! Save WRF soft cut-off in event record
        CALL WH_SetWrfEgmin(EminWrf)
! Generate photons and leptons in WRF: Q = (M,0,0,0)
        CALL GenYFS(MltFSR,KeyRem,IdGen,Q,aml,amn,delta,Emin,alfinv,
     &          thel,phil,ql,qn,nphot,photk,phsum,ys,zs,WtRad,MarTot)
        IF (WtRad .NE. 0d0) THEN 
           CALL WH_GetKeyCor(KeyRad)
           CALL WH_GetKeyQED(KeyQED)
*/ Get quarks data 
           CALL WH_GetIdQua(Idqu,Idaq)
           CALL WH_GetWrfMomQua(pqu,paq)
           IF (QW .GT. 0d0) THEN
              amd = Xmfe(-Idaq)
              amu = Xmfe( Idqu)
              DO k = 1,4
                 qd(k) = paq(k)
                 qu(k) = pqu(k)
              ENDDO
           ELSE
              amd = Xmfe( Idqu)
              amu = Xmfe(-Idaq)
              DO k = 1,4
                 qd(k) = pqu(k)
                 qu(k) = paq(k)
              ENDDO
           ENDIF
! Weight compensating for approximations in YFS formfactor
           CALL YWlRem(KeyRad,amd,amu,aM,aml,qd,qu,Q,ql,EminWrf,WtYff)
! Weight compensating for approximations in S-factor
           WtSfa = 1d0
           IF (KeyRad .GT. 2 .AND. nphot .GT. 0) THEN
              aml2 = aml**2
              amd2 = amd**2
              amu2 = amu**2
              DO i = 1, nphot
                 DO k = 1, 4
                    qk(k) = photk(i,k)
                 ENDDO
                 Sfin  = SWDecInt(amd,amu,aM,aml,qd,qu,Q,ql,qk)
                 Sfsr  = StPair(aM2,aml2,Q,ql,qk)
!--- Including ISR with dipole subtraction
                 IF (KeyQED .EQ. 2) THEN
                    DO k = 1,4
                       Qi(k) = Q(k) - qk(k)
                    ENDDO
                    Sfin = Sfin + StISRsub(Qi,qk)
                 ENDIF

                 WtSfa = WtSfa * (Sfin/Sfsr)
              ENDDO
           ENDIF
        ELSE
           WtYff = 0d0
           WtSfa = 0d0
        ENDIF
! Main weight
        Wt = WtRad * WtYff * WtSfa
*// Check on Wt - if not huge
        Huge = 1d6
        IF ( ABS(Wt) .LT. Huge) THEN
*//====>  To avoid NaNs
        ELSE
           NWtYFS = NWtYFS + 1
*//--- Print-outs if LevPrn > 0  
           IF (LevPrn .GE. 2) THEN
              write(*,*)'*********************************************'
              write(*,*)' >>>>>> KarwDec: Huge WtYFS !!! - reset to 0.'
              write(*,*)' No. NWtYFS = ',NWtYFS
              write(*,*)' WtYFS = ',Wt
              write(*,*)' WtRad,WtYff,WtSfa = ',WtRad,WtYff,WtSfa
              write(*,*)' amd,amu=',amd,amu
              write(*,*)' aM,aml,amn=',aM,aml,amn
              write(*,*)' qd=',qd
              write(*,*)' qu=',qu
              write(*,*)' Q =',Q
              write(*,*)' ql=',ql
              write(*,*)' qn=',qn
              write(*,*)' nphot=',nphot
              write(*,*)' EminWrf=',EminWrf
              write(*,*)'*********************************************'
           ENDIF
           Wt = 0d0
        ENDIF
*// Monitoring of main YFS weight 
        CALL gmonit(0,IdGen+69,Wt,1d0,0d0)

!WP: test monitoring of interference weight
!WP        IF (WtRad .NE. 0d0) THEN
!WP           Yscut = Yfscut(aM,aml,Q,ql,delta)
!WP           Yfsr  = YWldec(Q,ql,EminWrf)
!WP           WtYfsr = EXP(Yscut - Yfsr)
!WP           WtInt = WtYff * WtSfa * WtYfsr
!WP           IF ( ABS(WtInt) .GT. 10*Huge) THEN
!WP*/ Print-outs if LevPrn > 0  
!WP              IF (LevPrn .GE. 2) THEN
!WP                 write(*,*)'*******************************************'
!WP                 write(*,*)' >>>>>> KarwDec: Huge WtInt!!! - reset to 0'
!WP                 write(*,*)' WInt = ',WtInt
!WP                 write(*,*)' WtYfsr,WtYff,WtSfa = ',WtYfsr,WtYff,WtSfa
!WP                 write(*,*)' amd,amu=',amd,amu
!WP                 write(*,*)' aM,aml,amn=',aM,aml,amn
!WP                 write(*,*)' qd=',qd
!WP                 write(*,*)' qu=',qu
!WP                 write(*,*)' Q =',Q
!WP                 write(*,*)' ql=',ql
!WP                 write(*,*)' qn=',qn
!WP                 write(*,*)' nphot=',nphot
!WP                 write(*,*)' EminWrf=',EminWrf
!WP                 write(*,*)'*******************************************'
!WP              ENDIF
!WP              WtInt = 0d0
!WP           ENDIF
!WP        ELSE
!WP           WtInt = 0d0
!WP        ENDIF
!WP        CALL gf1(IdGen+110,WtInt,1d0)
!WP: End of monitoring interference weight

!=======================================================================
!======================== Final weights analysis =======================
!=======================================================================
      ELSE
*/ Print-outs if LevPrn > 0  
         CALL WH_GetLevPrn(LevPrn)
         IF (LevPrn.EQ.0) RETURN
!.........................Output window a...............................
        CALL gmonit(1,IdGen+60,AvMult,dumm2,EvTot)
        CALL gmonit(1,IdGen+61,awt61,Dwt61,EvTot)
        CALL gmonit(1,IdGen+62,awt62,Dwt62,EvTot)
        CALL gmonit(1,IdGen+63,awt63,Dwt63,EvTot)
        CALL gmonit(1,IdGen+69,awt69,Dwt69,EvTot)
! Get box-formats
        CALL WH_GetBxFmts(
     &          bxope,bxclo,bxtxt,bxl1i,bxl1l,bxl1f,bxl2f,bxl1g,bxl2g)
! General information on weights
        CALL WH_GetKeyPhy(Ivec)
        KeyWmp = Ivec(1)        ! W- or W+ or both
        WRITE(Nout,bxope)
        IF (KeyWmp .EQ. 1) THEN
           WRITE(Nout,bxtxt) ' ===> W- decay '
        ELSEIF (KeyWmp .EQ. 2) THEN
           WRITE(Nout,bxtxt) ' ===> W+ decay '
        ELSE
           WRITE(Nout,bxtxt) ' ===> W decay '
        ENDIF
        WRITE(Nout,bxtxt)'KarWdec Output - Window A'
        WRITE(Nout,bxtxt)'    Weight Statistics   '
        WRITE(Nout,bxl2f)awt69,dwt69,' General Weight    ','Wt    ','a1'
        WRITE(Nout,bxl1l)XevGen,     ' Generated Events  ','NevGen','a2'
        WRITE(Nout,bxl1f)AvMult,     ' Aver. ph. multi.  ','AvMult','a3'
        WRITE(Nout,bxl2f)Awt61,Dwt61,' Kinematics, smin  ','Wt1   ','a5'
        WRITE(Nout,bxl2f)Awt62,Dwt62,' Photons weight    ','Wt2   ','a6'
        WRITE(Nout,bxl2f)Awt63,Dwt63,' Phase-space weight','WtPS  ','a7'
        WRITE(Nout,bxclo)
! Specific details on mass weight rearrangenment, and rejection
        CALL gmonit(1,IdGen+64,AvMlt,Dwt64,EvTot)
!WP      CALL gmonit(1,IdGen+65,Awt65,Dwt65,EvTot)
!WP      CALL gmonit(1,IdGen+66,Awt66,Dwt66,EvTot)
!WP      CALL gmonit(2,IdGen+66,Evacc,EvNeg,EvOve)
!WP      CALL gmonit(3,IdGen+66,Evneg,EvZer,WwMax)
!WP        Nzer66 = EvZer
!WP        Ntot66 = EvTot
! Auxiliary print-outs
        WRITE(Nout,bxope)
        WRITE(Nout,bxtxt) 'KarWdec Output - Window B'
!WP        WRITE(Nout,bxtxt) '    On Photon Generator '
        WRITE(Nout,bxl1f) AvMlt,      ' raw ph. multipl.  ','    ','b1'
        WRITE(Nout,bxl1g) Emin,       ' photon Emin [GeV] ','    ','b2'
!WP        WRITE(Nout,bxtxt) '    On Mass Weights     '
!WP        WRITE(Nout,bxl1g) delta,      ' delta             ','    ','b3'
!WP      WRITE(Nout,bxl2f) Awt66,Dwt66,' Removal wgt Wtrem ','Wt6 ','b4'
!WP      WRITE(Nout,bxl1i) Ntot66,     ' No. of Raw Events ','    ','b5'
!WP      WRITE(Nout,bxl1i) Nzer66,     ' Wt6 = 0    Events ','    ','b6'
!WP      WRITE(Nout,bxl1i)MarTot,    ' Marked photons    ','MarTot','b7'
!WP      WRITE(Nout,bxl2f) Awt65,Dwt65,' Control wgt WtCtrl','Wt5 ','b8'
        WRITE(Nout,bxclo)
! histograms
*       CALL gprint(IdGen+20)
!WP: Interference weight
!WP        CALL gprint(IdGen+110)
      ENDIF
      END


      SUBROUTINE GenYFS(
     &     MltFSR,KeyRem,IdGen,Q,amle1,amle2,delta,Emin,alfinv,
     &     the,phi,q1,q2,nphot,phot,phsu,ygr,zet,WtWle,martot)
*     ************************************************************
!----------------------------------------------------------------------!
! Simulates multiphoton radiation in leptonic W-decays in the framework!
! of the YFS exponentiation (based on yfs_fin by S. Jadach).           !
! INPUT  :                                                             !
!     MltFSR  = normaly set to zero, otherwise enforces photon         !
!               multiplicity to be exactly equal MltFSR (special tests)!
!     KeyRem  = photon removal switch                                  !
!     Q     = 4-momentum of W                                        !
!     amle1,2 = masses of the final lepton pair (l,nu_l)               !
!     delta   = lower energy bound (dimensionless)                     !
!     emin    = lower energy bound (GeV) in LAB system                 !
!     alfinv  = 1/alpha_QED                                            !
!     the,phi = rotation angles for kinamatics calculation             !
! OUTPUT :                                                             !
!     q1,2    = final charged lepton and neutrino 4-momenta, resp.     !
!     nphot   = photon multiplicity                                    !
!     phot    = photon four momenta (GeV) in cms                       !
!     phsu    = sum of photon momenta                                  !
!     ygr,zet = Sudakov variables: ygr = 2(k.q1)/Q^2, zet = 2(k.Q)/Q^2 !
!     WtWle   = MC weight at this stage                                !
!     martot  = control variable, no of marked photons                 !
! OTHER:                                                               !
!     mk     = marks on photons CLOSE to lower energy bound            !
!     wt1    = the weight - phase space limits for very hard phot      !
!     wt2    = the weight - S-factors and YFS formfactor               !
!     wtm    = the list/matrix of mass weights                         !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek                    Knoxville, Aug. 1997  !
! Last update: 10.01.2006     by: W.P.                                 !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER (pi = 3.1415926535897932d0)
! Maximum number of radiative photons
      PARAMETER (Nmax = 100)
      DIMENSION phot(Nmax,4),phsu(4),mk(Nmax),wtm(Nmax)
      DIMENSION Q(4),q1(4),q2(4)
!WP      DIMENSION q1c(4),q2c(4),q1w(4)
      DIMENSION ygr(Nmax),zet(Nmax)
*
      DIMENSION rr(Nmax),xk(Nmax),cgx(Nmax),sgx(Nmax)
      DIMENSION dis0(Nmax)
      DIMENSION drvec(Nmax)
*-----------------------------------------------------------------------
      alf1 = 1d0/pi/alfinv
      Wt1  = 1d0
      Wt2  = 1d0
      WtPS = 1d0
!WP      WtCtrl = 1d0
!WP      WtRem = 1d0
*-----------------------------------------------------------------------
      svar = Q(4)**2 - Q(3)**2 - Q(2)**2 - Q(1)**2
      aM   = SQRT(svar)
      aml  = amle1
      aml2 = aml**2
      Enec = (svar + aml2)/2/aM   ! overvalued lepton energy for crude
      amc2 = (aml/Enec)**2 
      betc = (svar - aml2)/(svar + aml2)     
      DO i = 1,Nmax
         rr(i) = 0d0
         mk(i) = 0
      ENDDO
      DO k=1,4
         phsu(k) = 0d0
      ENDDO
      DO i = 1,Nmax
         ygr(i) = 0
         zet(i) = 0
         DO k=1,4
            phot(i,k) = 0d0
         ENDDO
      ENDDO
*-----------------------------------------------------------------------
*     Generate photon multiplicity, averg = average multiplicity (crude)
*-----------------------------------------------------------------------
      gamf2 = 2*alf1/betc*LOG(aM/aml)
      Averg = gamf2*dlog(1/delta)
    5 CONTINUE
      CALL WH_PoissG(Averg,Nmax,nphot,rr)
** This is for tests of program at fixed multiplicity (advanced users)
      IF((MltFSR .NE. 0) .AND. (nphot .NE. MltFSR)) GOTO 5
**
      IF(nphot .EQ. 0) THEN
         sprim = svar
*-----------------------------------------------------------------------
*     Reject events outside phase space
*-----------------------------------------------------------------------
         smini = (amle1 + amle2)**2
         IF(sprim .LT. smini) GOTO 900
      ELSE
*-----------------------------------------------------------------------
* Photon energy
         xsum = 0d0
         DO  i = 1,nphot
            xk(i) = delta**rr(i)
*--- Mark photons close to lower boundary for further control
            IF(xk(i) .LT. SQRT(10d0)*delta) mk(i) = 1
            xsum = xsum + xk(i)
         ENDDO
         IF (xsum .GE. 2d0) GOTO 900
*-----------------------------------------------------------------------
* Photon angular distributions in W rest frame with +z axis along q1
         CALL varran(drvec,nphot)
         DO i = 1,nphot
*-----------------------------------------------------------------------
*     simplified photon angular distribution,
*     E_l1 --> Enec, m**2/(kp)**2 dropped
*     cg=cos(theta) and sg=sin(theta) memorized to avoid rounding err.
!WP          CALL angble(amc2,dlt,cg,sg,dis0(i),dis1)
          CALL AngWle(aM,aml,amc2,betc,cg,sg,dis0(i))
*-----------------------------------------------------------------------
*     define photon momenta (in units of aM/2)
            phig = 2*pi*drvec(i)
            phot(i,1) = xk(i)*sg*cos(phig)
            phot(i,2) = xk(i)*sg*sin(phig)
            phot(i,3) = xk(i)*cg
            phot(i,4) = xk(i)
            DO k=1,4
               phsu(k) = phsu(k) + phot(i,k)
            ENDDO
            cgx(i) = cg
            sgx(i) = sg
         ENDDO
*-----------------------------------------------------------------------
*     Reject events with too much radiation (outside phase space)
*-----------------------------------------------------------------------
         xmk2  = phsu(4)**2-phsu(3)**2-phsu(2)**2-phsu(1)**2
         yy    = 1 - phsu(4) + xmk2/4
         sprim = svar*yy
         smini = (amle1 + amle2)**2
         IF(sprim .LT. smini) GOTO 900
*-----------------------------------------------------------------------
*     Rescale properly all photon momenta
*-----------------------------------------------------------------------
         ener = aM/2   
         DO  k = 1,4
            phsu(k) = phsu(k)*ener
            DO  i = 1,nphot
               phot(i,k) = phot(i,k)*ener
            ENDDO
         ENDDO
      ENDIF ! nphot
*-----------------------------------------------------------------------
*     Kinematics: final fermion 4-momenta in W rest-frame
*-----------------------------------------------------------------------
      amcr1 = amle1
      amcr2 = 0d0
      CALL KinWle(sprim,aM,amle1,amle2,phsu,q1,q2,bet1,WtKin) ! real
*     Reject events with too much radiation (outside phase space)
      IF (WtKin .EQ. 0d0) GOTO 900
!WP      CALL KinGho(aM,amcr1,amcr2,q1c,q2c,betc,etc1,etc2)      ! ghost
*... Save charged lepton 4-momentum in W-frame for later use
!WP      DO k = 1,4
!WP        q1w(k) = q1(k)
!WP      ENDDO
*-----------------------------------------------------------------------
* Weight correponding to phase-space integration over energy-momentum
* conservation delta-function
      WtPS = 2*q1(3)/( aM - phsu(4) + q1(4)/q1(3)*phsu(3) )
*-----------------------------------------------------------------------
*     Mass weight for theta distribution
*-----------------------------------------------------------------------
*     Mass weight compensates for s'->s and droping terms -m**2/(k.q)**2
*     Care is taken of machine rounding errors.
*     delt RECALCULATED out of angles sgx(i),cgx(i) 
*     with TRUE lepton energy, using EXACT formulas
      amd1 = (amle1/q1(4))**2
      DO i=1,nphot
         IF (cgx(i) .GT. 0d0) THEN
           delt = amd1/(1+bet1) + bet1*sgx(i)**2/(1+cgx(i))
         ELSE
           delt = 1d0 - bet1*cgx(i)
         ENDIF
         dist1  = 2/delt*(1d0 - amd1/2/delt - delt/2) 
         wtm(i) = dist1/dis0(i)
         IF(wtm(i) .LT.  1d-90) wtm(i)= 0d0
         Wt2 = Wt2*wtm(i)
*     Finaly define Sudakov variables
         ygr(i) = xk(i)*delt*q1(4)/aM
         zet(i) = xk(i)
      ENDDO
* The total phase space integral for crude x-section,
* Note that delta is a lower limit on x-variables in crude generation
      Fphs = 2*alf1/betc*LOG(aM/aml)*LOG(1/delta)
* YFS formfactor cut-off dependend part in W rest frame
* Note: The same cut-off delta as for "crude formfactor" is used here! 
      Fyfs  = Yfscut(aM,aml,Q,q1,delta)
      delb  = Fyfs + Fphs
      Wt2 = Wt2*exp(delb)
*-----------------------------------------------------------------------
* Rotations in WRF to get z-axis along the beams, with +z pointing in
* the initial quark direction.  
*-----------------------------------------------------------------------
      CALL RotWle(the,phi,q1,q2,nphot,phot,phsu)
*-----------------------------------------------------------------------
* Calculate YFS formfactor (cut-off dependent part) and mass weights
* Optionally removing photons below emin from the list
!WP      CALL RemWle(KeyRem,aM,amle1,emin,delta,alfinv,
!WP     &            Q,q1,q1c,q1w,phsu,phot,nphot,Wtm,WtRem,Wt2,WtCtrl)
      GOTO 1000
*-----------------------------------------------------------------------
* Event outside phase space (too much radiation)
 900  CONTINUE
      Wt1 = 0d0
      nphot = 0
*-----------------------------------------------------------------------
 1000 CONTINUE
      WtWle = Wt1*Wt2*WtPS
*-----------------------------------------------------------------------
* Monitoring weights and other control variables,
* Non-essential for the MC generation itself.
      CALL gmonit(0,IdGen+64 ,Averg,20d0,0d0)
!WP      CALL gmonit(0,IdGen+65 ,WtCtrl,1d0,0d0)
!WP      CALL gmonit(0,IdGen+66 ,WtRem,1d0,0d0)
* Monitoring weights
      CALL gmonit(0,IdGen+60,1d0*nphot,20d0,0d0)
      CALL gmonit(0,IdGen+61,Wt1,1d0,0d0)
      CALL gmonit(0,IdGen+62,Wt2,1d0,0d0)
      CALL gmonit(0,IdGen+63,WtPS,1d0,0d0)
* marked photons
      IF(nphot .GE. 1) THEN
         DO i = 1,nphot
            ul = LOG10(phot(i,4)/emin)
            IF (mk(i) .EQ. 1)  CALL gf1(IdGen+20,   ul,1d0)
            IF (mk(i) .EQ. 1)  martot=martot+1
         ENDDO
      ENDIF
      END

      SUBROUTINE AngWle(aM,aml,amx2,beta,costhg,sinthg,dist)
*     *******************************************************
!----------------------------------------------------------------------!
! This routine generates photon angular distribution in leptonic       !
! W-decays in the rest frame of W.                                     !
! The distribution is the S-factor without mass terms.                 !
! INPUT: aM,aml - W invariant mass and leton mass                      !
!        amx2 = aml**2/El**2, where El - lepton energy                 !
!        beta - lepton relativistic velocity                           !
! OUTPUT:                                                              !
!     del = 1-beta*cos(theta)                                          !
!     costhg, sinthg - cos and sin of the photon polar angle           !
!                      with respect to charged lepton direction        !
!     dist - generated distribution                                    !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek                         CERN, July 2002  !
! Last update: 12.12.2002           by: W.P.                           !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DIMENSION rn(1)
! Generation of a random number
      CALL varran(rn,1)
! Generation of cos(theta_gamma)
      R2 = 2*rn(1)
      del = (1d0 + beta) / (aM/aml)**R2             ! 1 - beta*costhg
      costhg = (1d0 - del)/beta 
      sinthg = SQRT( del*(2d0 - del) - amx2*costhg**2  )      
      dist = 2d0/del
      END

      SUBROUTINE angble(am2,del,costhg,sinthg,dist0,dist1)
*     ****************************************************
*WP: OLD !
!----------------------------------------------------------------------!
! This routine generates photon angular distribution in leptonic       !
! W-decays in the rest frame of W.                                     !
! The distribution is the S-factor without mass terms.                 !
! This is modified version of routine angbre from YFS3.                !
!                                                                      !
! INPUT:                                                               !
!     am2 = massl**2/El**2, where massl, El are lepton mass and energy !
! OUTPUT:                                                              !
!     del = 1-beta*cos(theta)                                          !
!     costhg, sinthg - cos and sin of the photon polar angle           !
!     angle with respect to fermions direction                         !
!     dist0 - distribution generated, without m**2/(kp)**2 terms       !
!     dist1 - distribution with m**2/(kp)**2 terms                     !
!----------------------------------------------------------------------!
! Modified by: Wieslaw Placzek                   Knoxville, July 1997  !
! Last update:  6.08.1997           by: W.P.                           !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DIMENSION rn(10)
! Generation of 1-beta*costhg
      CALL varran(rn,1)
      beta = SQRT(1d0 - am2) 
      eps  = am2/(1d0 + beta)                       != 1-beta
      del1 = (2d0 - eps)*( eps/(2d0 - eps) )**rn(1) != 1-beta*costhg
      del2 = 2d0 - del1                             != 1+beta*costhg
      del  = del1 
! Calculation of sin and cos theta from internal variables
      costhg = (del2 - del1)/(2*beta)               ! exact
      sinthg = SQRT(del1*del2 - am2*costhg**2)      ! exact
      dist0 = 2/del
      dist1 = 2/del*(1d0 - am2/2/del - 0.5*del) 
      END

      SUBROUTINE KinWle(sprim,aM,aml1,aml2,phsum,q1,q2,bet1,WtKin)
*     *************************************************************
!----------------------------------------------------------------------!
! This routine provides 4-momenta of final state leptons in radiative  !
! W-decays in the W rest frame with +z axis along the momentum q1,     !
! after photons 4-momenta generation is performed.                     !
!                                                                      !
! INPUT: sprim = (q1 + q2)**2                                          !
!        aM -- mass of W                                               !   
!        aml1,aml2 -- masses of final state leptons: l and nu_l, resp. !    
!        phsum -- sum of photons 4-momenta                             !
! OUPUT: q1,q2 -- 4-momenta of leptons: l and nu_l, resp.              !
!        bet1  -- velocity of the charged lepton                       !
!        Wtkin -- kinematics weight                                    !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek                    Knoxville, Aug. 1997  !
! Last update: 02.12.1998           by: W.P.                           !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DIMENSION phsum(4),q1(4),q2(4)
!
      WtKin = 0d0
      aml1s = aml1**2
      aml2s = aml2**2
      E12  = aM - phsum(4)
      phs3 = phsum(3)
      eph3 = (E12 - phs3)*(E12 + phs3)
      rsp  = sprim + aml1s - aml2s
      dlt  = rsp**2 - 4*aml1s*eph3
! Outside phase space
      IF (dlt .LE. 0d0) RETURN
      q1m  = ( E12*SQRT(dlt) - rsp*phs3 )/2/eph3
! Outside phase space
      IF (q1m. LE. 0d0) RETURN
      E1   = SQRT(q1m**2 + aml1s)
! Charged lepton 4-momentum
      q1(1) = 0d0
      q1(2) = 0d0
      q1(3) = q1m
      q1(4) = E1
! Neutrino 4-momentum
      q2(1) = -phsum(1)
      q2(2) = -phsum(2)
      q2(3) = -phsum(3) - q1m
      q2(4) = E12 - E1
! Velocity of charged lepton
      bet1 = q1m/E1
! Kinematics weight 
      WtKin = 1d0
      END

      SUBROUTINE KinGho(cmsene,am1,am2,p1,p2,beta,eta1,eta2)
*     ******************************************************
* For CMS energy = cmsene it defines two "decay" momenta p1,p2
* in their rest frame, along z-axix
*     ***********************************
!WP: This routine comes from yfs3ff version 3.4.0 (August 1997)
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      DIMENSION p1(*),p2(*)
*
      ener  =  cmsene/2d0
      svar  =  cmsene**2
      beta  =  SQRT( (svar-(am1-am2)**2)
     $              *(svar-(am1+am2)**2) )/svar
      eta1=    (svar+am1**2-am2**2)/svar
      eta2  =  (svar-am1**2+am2**2)/svar
*----------------------------------------------
      p1(1)  =  0d0
      p1(2)  =  0d0
      p1(3)  =  ener*beta
      p1(4)  =  ener*eta1
*----------------------------------------------
      p2(1)  =  0d0
      p2(2)  =  0d0
      p2(3)  = -ener*beta
      p2(4)  =  ener*eta2
      END

      SUBROUTINE RotWle(the,phi,q1,q2,nphot,phot,phsu)
*     *************************************************
!----------------------------------------------------------------------!
! Random Euler rotations in the W rest frame (WRF).                    !
! INPUT: the, phi - polar and azimuthal (Euler) angles                 !
!        nphot - numer of photons                                      !
! Others - 4-momenta before and after transformations                  !
!----------------------------------------------------------------------!
! Written: Wieslaw Placzek                            CERN, July 2002  !
! Last update: 21.07.2002           by: W.P.                           !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER( pi = 3.1415926535897932d0)
      DIMENSION Q(4),phot(100,4),phsu(4),q1(4),q2(4)
      DIMENSION ph(4)
!
! Rotations in W rest frame
      CALL Rot2le(the,phi,q1,q1)
      CALL Rot2le(the,phi,q2,q2)
      CALL Rot2le(the,phi,phsu,phsu)
      DO i = 1,nphot
         DO k = 1,4
            ph(k) = phot(i,k)
         ENDDO
         CALL Rot2le(the,phi,ph,ph)
         DO k=1,4
            phot(i,k) = ph(k)
         ENDDO
      ENDDO
      END

      SUBROUTINE Rot2le(the,phi,pvec,qvec)
*     ****************************************
! Euler rotation and boosts for leptonic W-decays
      IMPLICIT REAL*8(a-h,o-z)
      REAL*8 pvec(4),qvec(4),Q(4)
!
      CALL Rxtod2(the,pvec,qvec)       ! rotation z-x
      CALL Rxtod3(phi,pvec,qvec)       ! rotation x-y
      END

      SUBROUTINE RemWle(KeyRem,aM,aml,emin,delta,alfinv,
     $           Q,ql,qlc,qlw,phsu,phot,nphot,wtmas,wtrem,wtm3a,wctrl)
*     *****************************************************************
!----------------------------------------------------------------------!
! This is modified version of the routine piatek from YFS3 - adapted   !
! for leptonic W-decays.                                               !
! Written CERN, piatek, 22 Sept. 1989  (S.J.)                          !
! This routine calculates YFS form-factor and optionaly                !
! removes some of soft photons (below emin in CMS), calculating        !
! compensating weight.                                                 !
! Note the action of this routine is not Loretnz invariant !!!!        !
! Input:                                                               !
!     KeyRem   = 0, NO removal of photons below emin                   !
!              = 1, with removal of photons below emin                 !
!     aM,aml  = masses of W and charged lepton                         !
!     emin     = emin minimum-energy of photons to be kept (GeV)       !
!     delta    = infrared cut-off in generation (dimensionless)        !
!     alfinv   = 1/alpha_QED                                           !
!     Q,ql    = 4-momenta of W and charged lepton                      !
!     qlc      = ghost lepton 4-momentum in crude S-factor             !
!     qlw      = charged lepton 4-momentum in W rest frame             !
!     phsu     = sum of photon 4-momenta                               !
!     phot     = list of photon 4-momenta                              !
!     nphot    = photon multiplicity                                   !
!     wtmas    = list of mass-weights for all photons                  !
! Output:                                                              !
!     wtrem    = 1,  for KeyRem=0                                      !
!              = mass-weight of removed photons, for KeyRem=1          !
!     wtm3a    = mass-weight for all photons, see comments below       !
!     wctrl    = control-weight for remowed photons                    !
!----------------------------------------------------------------------!
! Modified by: Wieslaw Placzek                   Knoxville, Aug. 1997  !
! Last update: 19.07.2002 (CERN)    by: W.P.                           !
!----------------------------------------------------------------------!
      IMPLICIT DOUBLE PRECISION (a-h,o-z)
      PARAMETER(pi=3.1415926535897932d0)
      DIMENSION  Q(4),ql(4),qlc(4),qlw(4),phsu(4),phot(100,4)
      DIMENSION  wtmas(100)
!
      alf1 = 1d0/pi/alfinv
      svar = aM**2
      wtm1=1d0
      wtm2=1d0
* Mass weight below and above emin calculated separately
      DO i=1,nphot
         IF(phot(i,4) .LT. emin) THEN
            wtm1=wtm1*wtmas(i)
            IF(wtm1 .LE. 1d-90) wtm1=0d0
         ELSE
            wtm2=wtm2*wtmas(i)
            IF(wtm2 .LE. 1d-90) wtm2=0d0
         ENDIF
      ENDDO
* The total phase space integral for crude x-section,
* Note that delta is a lower limit on x-variables in crude generation
      betc  = (svar - aml**2)/(svar + aml**2)
      fphs  =  2*alf1/betc*LOG(aM/aml)*LOG(1/delta)
* YFS formfactor cut-off dependend part in W rest frame
* Note: The same cut-off delta as for "crude formfactor" is used here! 
      fyfs  = Yfscut(aM,aml,Q,ql,delta)
      delb  = fyfs + fphs
* The average mass-weight for removed photon = exp(delb2)
* It can be calculated analytically as a  ratio of YFS formfactors
* On the other hand, it is checked by MC, see control weight wctrl
      delb2 = DelB2l(emin,delta,aM,aml,Q,ql,qlc,qlw)
* Control weight - its average should be =1 within statist. error
      wctrl =wtm1*exp(-delb2)
      IF(KeyRem .EQ. 0) THEN
         IF(abs(delb) .GT. 100d0) WRITE(*,*) ' delb= ',delb
         wtrem = 1d0
         wtm3a = wtm1*wtm2*exp(delb)
      ELSE
* Optional removal of photons below emin from the record
* in such a case wtm3a includes exp(delb2)= <wt3> for removed ph.
         nph = nphot
         DO j = nphot,1,-1
            IF(phot(j,4) .LT. emin) THEN
               DO i = j+1,nph
                  DO k = 1,4
                     phot(i-1,k) = phot(i,k)
                  ENDDO
               ENDDO
               nph = nph-1
            ENDIF
         ENDDO
         DO j = nph+1,nphot
            DO k=1,4
               phot(j,k) = 0d0
            ENDDO
         ENDDO
         nphot = nph
* wtmas includes here average weight of removed photons exp(delb2)
         wtrem = wtm1
         wtm3a = wtm2*exp(delb+delb2)
         IF (wtm3a.LT.1d2) THEN
c           print*,'wtm3a=',wtm3a
         ELSE
           print*,'wtm3a,delb2=',wtm3a,delb2
           print*,'aM,aml=',aM,aml
           print*,'Q =',Q
           print*,'ql=',ql
           print*,'qlc=',qlc
           print*,'qlw=',qlw
           print*,'nphot=',nphot
           wtm3a = 0d0
         ENDIF
      ENDIF
      END ! RemWle

      FUNCTION DelB2l(emin,delta,aM,aml,Q,ql,rl,qlw)
*     ************************************************
!----------------------------------------------------------------------!
! This is a modified version of the routine delb2f from YFS3 - adapted !
! for leptonic W-decays.                                               !
! Exact formula for delb2 in the compensating weight exp(delb2)        !
! necessary in the case of erasing soft photons below emin,            !
! version for the case of finite mass.                                 !
! It is a linear combination of 4 Breal function:                      !
! 1 exact and 1 crude, for 2 boundaries (LAB and W rest frame).        !
!                                                                      !
! INPUT: emin,delta - soft photon cut-offs in LAB and W rest frame     !
!        aM,aml    - masses of W and charged lepton                    !
!        Q,ql,rl   - 4-momenta of W, lepton and ghost                  !
!        qlw        - 4-momentum of lepton in W rest frame             !
!                                                                      !
! Remark on rl:                                                        !
! It is ghost-fermion 4-momentum in the definition of the crude        !
! distribution, i.e. truncated S-factor is 2*Q*rl/(k*Q)/(k*rl)         ! 
! as generated in the crude MC.                                        !
!----------------------------------------------------------------------!
! Modified by: Wieslaw Placzek                   Knoxville, Aug. 1997  !
! Last update: 08.08.1997           by: W.P.                           !
!----------------------------------------------------------------------!
      IMPLICIT REAL*8(a-h,o-z)
      REAL*8 Q(4),ql(4),rl(4),qlw(4)
!
! Lepton energy in W-frame
      Elc = (aM  + aml**2/aM)/2       ! ghost
      Ele = qlw(4)                    ! real 
      eming = delta*aM/2              ! cut-off in W-frame
      amgam =1d-20                    ! dummy photon mass
********* various input parameters for Btildc and Btildf
      Qql  = Q(4)*ql(4) - Q(3)*ql(3) - Q(2)*ql(2) - Q(1)*ql(1)
      Qrl  = Q(4)*rl(4) - Q(3)*rl(3) - Q(2)*rl(2) - Q(1)*rl(1)
********* CMS boundary, difference crude-exact
      dcms2c = Btildc(Qrl, Q(4),rl(4), aM,aml, emin,amgam) !crude
      dcms2f = Btildf(Qql, Q(4),ql(4), aM,aml, emin,amgam) !exact
********* WMS boundary, difference crude-exact
      dqms2c = Btildc(Qrl, aM,Elc, aM,aml, eming,amgam) !crude
      dqms2f = Btildf(Qql, aM,Ele, aM,aml, eming,amgam) !exact
********* QMS - CMS, result should be negative ( delta<<epsilon )
      delb2l = (dqms2c-dqms2f) - (dcms2c-dcms2f)
      IF (delb2l.LE.1d3) THEN

      ELSE
        print*,'Qrl,Qql=',Qrl,Qql
        print*,'rl=',rl
        print*,'ql=',ql
        print*,'dcms2c,dcms2f=',dcms2c,dcms2f
        print*,'dqms2c,dqms2f=',dqms2c,dqms2f
      ENDIF
      END

      SUBROUTINE YWlRem(KeyRad,amd,amu,amW,aml,qd,qu,Q,ql,Emin,WtYff)
!     ***************************************************************
!----------------------------------------------------------------------!
! A Weight compensating for simplified YFS formfactor                  !
! used at the generation level for radiative W-decays.                 !
!                                                                      !
! INPUT: KeyRad -- radiative corrections switch                        ! 
!        amd,amu -- masses of initial quarks d (s, b) and u (c, t)     !
!        amW,aml  -- masses of W (invariant) and final-state lepton    !
!        qd,qu   -- 4-momenta of initial quarks d (s, b) and u (c, t)  !
!        Q,ql    -- 4-momenta of W and lepton, resp.                   !
!        Emin    -- soft photon energy cut-off [in GeV]                !
! OUPUT: WtYff   -- compensation weight                                !
!----------------------------------------------------------------------!
! Written by: Wieslaw Placzek                        CERN,  July 2002  !
! Last update: 24.10.2008           by: W.P. (GGI, Florence)           !
!----------------------------------------------------------------------!
      IMPLICIT REAL*8 (A-H,O-Z)
      PARAMETER( pi = 3.1415926535897932d0, alfinv=137.03599911d0)
      PARAMETER( alfpi=  1/pi/alfinv, alfa= 1d0/alfinv)
      DIMENSION qd(4),qu(4),Q(4),ql(4)
!
! ... YFS formfactor weight
      epsg  = 2*Emin/amW
      Yscut = Yfscut(amW,aml,Q,ql,epsg)
      IF (KeyRad .LE. 2) THEN
! ... FSR only
         Yfull = YWldec(Q,ql,Emin)
      ELSE
! ... FSR + Interference
         Yfull = YWDecInt(amd,amu,amW,aml,qd,qu,Q,ql,Emin)
!--- Including ISR with dipole subraction
         CALL WH_GetKeyQED(KeyQED)
         IF (KeyQED .EQ. 2) THEN
            sqq = (qd(4) + qu(4))**2 - (qd(3) + qu(3))**2
     &          - (qd(2) + qu(2))**2 - (qd(1) + qu(1))**2
            Yfull =  Yfull + WH_YISRsDip(sqq,Emin)
         ENDIF
      ENDIF
      Yrem = Yfull - Yscut
      WtY   = EXP(Yrem)
! Total weight
      WtYff = WtY
      END

      SUBROUTINE WboostL(s,Q,q1,q2,nphot,phot,phsu,ql1,ql2,photl,phsul)
*     *****************************************************************
!----------------------------------------------------------------------!
! Lorentz transforms from W rest frame to LAB: q1, q2, phot, phsu      !
! with random Euler rotation in the intermediate Q frame (W frame)     !
! INPUT: s - W invariant mass squared                                  !
!        Q - W 4-momentum                                              !
!        nphot - numer of photons                                      !
! Others - 4-momenta before and after transformations                  !
!----------------------------------------------------------------------!
! Written: Wieslaw Placzek                            CERN, July 2002  !
! Last update: 21.07.2002           by: W.P.                           !
!----------------------------------------------------------------------!
      DIMENSION Q(4),q1(4),q2(4),phot(100,4),phsu(4)
      DIMENSION ql1(4),ql2(4),photl(100,4),phsul(4)
      DIMENSION ph(4)
! Boosts from W rest frame to LAB, if necessary
      betW2 = 1d0 - s/Q(4)**2
      IF (betW2 .GT. 1d-10) THEN
         CALL BostdQ(-1,Q,q1,ql1)
         CALL BostdQ(-1,Q,q2,ql2)
         CALL BostdQ(-1,Q,phsu,phsul)
         DO i=1,nphot
            DO k=1,4
               ph(k)=phot(i,k)
            ENDDO
            CALL BostdQ(-1,Q,ph,ph)
            DO k=1,4
               photl(i,k)= ph(k)
            ENDDO
         ENDDO
      ENDIF
      END

      SUBROUTINE WH_PoissG(Averg,NMax,Mult,Rr)
*///////////////////////////////////////////////////////////////////////
*// This generates photon multipl. NPHOT according to Poisson distr.  //
*// INPUT:  Averg = Average multiplicity                              //
*//         NMax  = Maximum multiplicity                              //
*// OUTPUT: Mult = Generated multiplicity                             //
*//         Rr(1:100) - List of ordered uniform random numbers,       //
*//         a byproduct result, to be optionally used for some other  //
*//         purpose (e.g. generation of photon energies).             // 
*//-------------------------------------------------------------------//
*// Written by S. Jadach,                                             //
*// Last update: 13.08.2002         by: WP                            //
*///////////////////////////////////////////////////////////////////////
      IMPLICIT DOUBLE PRECISION (A-H,O-Z) 
      PARAMETER ( Nlim = 100 )
      DIMENSION Rr(*)                    
      DIMENSION Drvec(1)
      DATA Nfail /0/                      
      SAVE Nfail
*
 50   Nn = 0                               
      Sum = 0D0                            
      DO It = 1,NMax                   
         CALL VarRan(Drvec,1)
         Rn = Drvec(1)
         Y = LOG(Rn)                         
         Sum = Sum + Y                          
         Nn = Nn + 1                            
         Rr(Nn) = Sum/(-Averg)                
         IF(Sum .LT. -Averg) GOTO 100         
      ENDDO
      Nfail = Nfail + 1                      
      IF (Nfail .GT. Nlim) GOTO 900          
      GOTO 50                            
 100  Mult = Nn - 1                         
      RETURN                       
 900  CONTINUE
      CALL WH_GetIO(Nint,Nout)      
      WRITE(Nout,*) ' >>> WH_PoissG: TOO SMALL NMax = ',NMax
      STOP                               
      END                                
