*//////////////////////////////////////////////////////////////////////////////
*//                                                                          //
*//  WINDEC is a wrapper for the MC event generator WINHAC for the process:  //
*//                      W --> l + nu + photons                              //
*//       with the O(alpha) Yennie-Frautschi-Suura exponentiation            //
*//                                                                          //
*//                    by W. Placzek and S. Jadach                           //
*//                                                                          //
*//                       Version of 09 May 2003                             //
*//                                                                          //
*//    Last correction: 05.08.2008 by WP.                                    //
*// =========================================================================//
*//                                                                          //
*// How to use it:                                                           //
*// --------------                                                           //
*//                                                                          //
*// 1. Set-up input values and send them to WINDEC with the help of the      //
*//    routine WINDEC_SetXpar, for details on input values see the routine   //
*//    WINDEC_Defaults (below in this file).                                 //
*//                                                                          //
*// 2. Initialize WINDEC by calling:                                         //
*//                                                                          //
*//       CALL WINDEC_Initialize                                             //
*//                                                                          //
*// 3. Generate an event of single W-boson production in hadronic collisions //
*//    with your preferred MC program.                                       //
*//                                                                          //
*// 4. For generated flavours (PDG codes) and four-momenta of a              //
*//    quark (Idqu, pqu) and an anti-quark (Idaq, paq) call:                 //
*//                                                                          //
*//       INTEGER           Idqu,Idaq,Idle,Idnu,Ngam                         //
*//       DOUBLE PRECISION  pqu(4),paq(4),qle(4),qnu(4),qgam(100,4)          //
*//                                                                          //
*//       CALL WINDEC_Make(Idqu,Idaq,pqu,paq,Idle,Idnu,qle,qnu,Ngam,qgam)    //
*//                                                                          //
*//    see routine WINDEC_Make for details (below in this file).             //
*//                                                                          //
*//    For each event, WINDEC will provide the flavours (PDG codes)          //
*//    (Idle, Idnu) and four-momenta (qle, qnu) of the final-state charged   //
*//    lepton and neutrino, respectively, as well as the number Ngam and     //
*//    four-momenta qgam of radiative photons.                               //
*//    All the final-state four-momenta are given in the same frame as the   //
*//    four-momenta of the incoming quarks, which can be defined in any      //
*//    reference frame (decay products are always tranformed to this frame). //
*//    Alternatively, they can be accesed through the following routines:    //
*//                                                                          //
*//       INTEGER           Idle,Idnu,Ngam                                   //
*//       DOUBLE PRECISION  qW(4),qle(4),qnu(4),qgam(100,4)                  //
*//                                                                          //
*//       CALL WH_GetLabMomW(qW)                                             //
*//       CALL WH_GetIdLep(Idle,Idnu)                                        //
*//       CALL WH_GetLabMomLep(qle,qnu)                                      //
*//       CALL WH_GetLabMomPho(Ngam,qgam)                                    //
*//                                                                          //
*//    In the weighted-event mode one can access the main event weight by:   //
*//                                                                          //
*//       DOUBLE PRECISION  WtMain                                           //
*//                                                                          //
*//       CALL WH_GetWtMain(WtMain)                                          //
*//                                                                          //
*// 5. After event genaration is completed, call:                            //
*//                                                                          //
*//       CALL WINDEC_Finalize                                               //
*//                                                                          //
*//    to calculate and print-out some final statistics.                     //
*//                                                                          //
*// NOTE: For the most common use, the unweighted-event mode (KeyWgt = 0)    //
*//       is recommended.                                                    //
*//                                                                          //
*//////////////////////////////////////////////////////////////////////////////
      SUBROUTINE  WINDEC_Preset
*     *************************
      IMPLICIT NONE
      INCLUDE 'windec.h'
*//////////////////////////////////////////////////////////////////////////////
*//   Class  defaults initialization, ...
*//////////////////////////////////////////////////////////////////////////////

      IF(m_Magic .EQ. 0 ) RETURN
      m_Magic = 0

      m_Ninp = 15
      m_Nout = 16
      CALL WINDEC_Defaults
      END

      SUBROUTINE  WINDEC_Defaults
*     ***************************
      IMPLICIT NONE
      INCLUDE 'windec.h'
      INTEGER i
      DO i=1,10000
         m_Xpar(i)=0d0
      ENDDO
*//////////////////////////////////////////////////////////////////////////////
*//                                                                          //
*//               DEFAULT   Input data for WINHAC                            //
*//                          (based on PDG 2002)                             //
*//                                                                          //
*//////////////////////////////////////////////////////////////////////////////
*BeginX
*<-i><----data-----><-------------------comments------------------------------>
      m_Xpar(    1)=       80.398d0 !  Ecm    =xpar( 1) ! CMS total energy in GeV
      m_Xpar(    2)=       80.398d0 !  aMW    =xpar( 2) ! W mass 
      m_Xpar(    3)=       -2.141d0 !  GamW   =xpar( 3) ! W width, if GamW<0 it is RECALCULATED 
*                                     !          inside the program
      m_Xpar(    4)=      91.1876d0 !  aMZ    =xpar( 4) ! Z mass   
      m_Xpar(    5)=       2.4952d0 !  GamZ   =xpar( 5) ! Z width  
      m_Xpar(    6)=      -0.2222d0 !  sW2    =xpar( 6) ! sin^2(theta_W), if sW2 < 0 it is      
*                                     ! RECALCULATED in the program
      m_Xpar(    7)=          150D0 !  aMH    =xpar( 7) ! Higgs mass  
      m_Xpar(    8)=          1.0D0 !  aGH    =xpar( 8) ! Higgs width 
      m_Xpar(    9)=           1d-3 !  EgMin  =xpar( 9) ! Soft-photon cut-off in GeV
      m_Xpar(   21)=   0.93827200d0 !  ampro  =xpar(21) ! proton mass
      m_Xpar(   22)= 137.03599976d0 !  alfinv =xpar(22) ! 1/alpha_QED 
      m_Xpar(   23)=     1.16639d-5 !  Gmu    =xpar(23) ! Fermi Constant [GeV^{-2}] 
      m_Xpar(   24)=       0.1185d0 !  alpha_s=xpar(24) ! QCD coupling constant at M_W
      m_Xpar(   25)=   389.379292d3 !  GeV2nb =xpar(25) ! Conversion constant: GeV^{-2} -> nb
      m_Xpar(   31)=           1d-6 !  x_min  =xpar(31) ! minimum of quarks x's
      m_Xpar(   32)=            1d0 !  x_max  =xpar(32) ! maximum of quarks x's
      m_Xpar(   33)=            1d0 !  Q2_min =xpar(33) ! minimum of quarks Q^2 [GeV^2]
      m_Xpar(   51)=            2d0 !  WtMax  =xpar(51) ! maximum weight for rejection 
*/////////////////////////////////////////////////////////////////////////
*//    Values of the CKM matrix elements from 2000 PDG Review:          //
*//  Eq. (11.3) for mean values of given ranges for s_12, s_23, s_13    //
*//     (only real parts of the matrix elements are taken)              //
*/////////////////////////////////////////////////////////////////////////
      m_Xpar(  101)=      0.97483d0 !  VCKM(1,1) =xpar(111) ! V_ud  real part
      m_Xpar(  102)=      0.22290d0 !  VCKM(1,2) =xpar(112) ! V_us  real part
      m_Xpar(  103)=      0.00360d0 !  VCKM(1,3) =xpar(113) ! V_ub  real part
      m_Xpar(  104)=     -0.22286d0 !  VCKM(2,1) =xpar(114) ! V_cd  real part
      m_Xpar(  105)=      0.97398d0 !  VCKM(2,2) =xpar(115) ! V_cs  real part
      m_Xpar(  106)=      0.04120d0 !  VCKM(2,3) =xpar(116) ! V_cb  real part
      m_Xpar(  107)=      0.00568d0 !  VCKM(3,1) =xpar(117) ! V_td  real part
      m_Xpar(  108)=     -0.04097d0 !  VCKM(3,2) =xpar(118) ! V_ts  real part
      m_Xpar(  109)=      0.99914d0 !  VCKM(3,3) =xpar(119) ! V_tb  real part
*       111-119 reserved for imaginary parts
*******************************************************************************
*     PROPERTIES of Quarks and Leptons
*     i-th position  in xpar(i) according to i = 500+10*KFlavour +j
****** d-quark
      m_Xpar(  511)=              1 !       KFlavour
      m_Xpar(  512)=              3 !       NColor
      m_Xpar(  513)=             -1 !       3*Q   =3*charge
      m_Xpar(  514)=             -1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  515)=              0 !       2*helicity, 0 for unpolarized
      m_Xpar(  516)=      0.00675d0 !       mass [GeV] (3-9 MeV in PDG)
      m_Xpar(  517)=          3.5d0 !       WtMax Maximum weight for rejection d-quark
****** u-quark
      m_Xpar(  521)=              2 !       KFlavour
      m_Xpar(  522)=              3 !       NColor
      m_Xpar(  523)=              2 !       3*Q   =3*charge
      m_Xpar(  524)=              1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  525)=              0 !       2*helicity, 0 for unpolarized
      m_Xpar(  526)=        0.003d0 !       mass [GeV] (1-5 MeV in PDG)
      m_Xpar(  527)=          3.5d0 !       WtMax Maximum weight for rejection u-quark
*indx+++++++++++++++ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0
****** s-quark
      m_Xpar(  531)=              3 !       KFlavour
      m_Xpar(  532)=              3 !       NColor
      m_Xpar(  533)=             -1 !       3*Q   =3*charge
      m_Xpar(  534)=             -1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  535)=              0 !       2*helicity, 0 for unpolarized
      m_Xpar(  536)=       0.1175d0 !       mass [GeV] (75-170 MeV in PDG)
      m_Xpar(  537)=          3.5d0 !       WtMax Maximum weight for rejection s-quark
****** c-quark
      m_Xpar(  541)=              4 !       KFlavour
      m_Xpar(  542)=              3 !       NColor
      m_Xpar(  543)=              2 !       3*Q   =3*charge
      m_Xpar(  544)=              1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  545)=              0 !       2*helicity, 0 for unpolarized
      m_Xpar(  546)=         1.20d0 !       mass [GeV] (1.15-1.35 GeV in PDG)
      m_Xpar(  547)=          3.5d0 !       WtMax Maximum weight for rejection c-quark
*indx+++++++++++++++ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0
****** b-quark
      m_Xpar(  551)=              5 !       KFlavour
      m_Xpar(  552)=              3 !       NColor
      m_Xpar(  553)=             -1 !       3*Q   =3*charge
      m_Xpar(  554)=             -1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  555)=              0 !       2*helicity, 0 for unpolarized
      m_Xpar(  556)=         4.25d0 !       mass [GeV] (4.0-4.4 GeV in PDG)
      m_Xpar(  557)=          3.5d0 !       WtMax Maximum weight for rejection b-quark
****** t-quark
      m_Xpar(  561)=              6 !       KFlavour
      m_Xpar(  562)=              3 !       NColor
      m_Xpar(  563)=              2 !       3*Q   =3*charge
      m_Xpar(  564)=              1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  565)=              0 !       2*helicity, 0 for unpolarized
      m_Xpar(  566)=        174.3d0 !       mass [GeV] (174.3 GeV in PDG)
      m_Xpar(  567)=          3.5d0 !       WtMax Maximum weight for rejection t-quark
*indx+++++++++++++++ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0
****** electron
      m_Xpar(  611)=             11 !       KFlavour
      m_Xpar(  612)=              1 !       NColor
      m_Xpar(  613)=             -3 !       3*Q   =3*charge
      m_Xpar(  614)=             -1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  615)=              0 !       2*helicity
      m_Xpar(  616)= 0.510998902d-3 !       mass [GeV]   (0.51099907 MeV)
      m_Xpar(  617)=          3.5d0 !       WtMax Maximum weight for rejection electron
****** neutrino electron
      m_Xpar(  621)=             12 !       KFlavour
      m_Xpar(  622)=              1 !       NColor
      m_Xpar(  623)=              0 !       3*Q   =3*charge
      m_Xpar(  624)=              1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  625)=              0 !       2*helicity
      m_Xpar(  626)=            0d0 !       mass [GeV]
      m_Xpar(  627)=          3.5d0 !       WtMax Maximum weight for rejection neutrino electron
*indx+++++++++++++++ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0
****** muon
      m_Xpar(  631)=             13 !       KFlavour
      m_Xpar(  632)=              1 !       NColor
      m_Xpar(  633)=             -3 !       3*Q   =3*charge
      m_Xpar(  634)=             -1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  635)=              0 !       2*helicity
      m_Xpar(  636)=  0.105658357d0 !       mass [GeV]   (0.105658389 GeV)
      m_Xpar(  637)=          3.5d0 !       WtMax Maximum weight for rejection muon
****** neutrino muon
      m_Xpar(  641)=             14 !       KFlavour
      m_Xpar(  642)=              1 !       NColor
      m_Xpar(  643)=              0 !       3*Q   =3*charge
      m_Xpar(  644)=              1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  645)=              0 !       2*helicity
      m_Xpar(  646)=            0d0 !       mass [GeV]
      m_Xpar(  647)=          3.5d0 !       WtMax Maximum weight for rejection neutrino muon
*indx+++++++++++++++ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0ccccccccc0
****** tau
      m_Xpar(  651)=             15 !       KFlavour
      m_Xpar(  652)=              1 !       NColor
      m_Xpar(  653)=             -3 !       3*Q   =3*charge
      m_Xpar(  654)=             -1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  655)=              0 !       2*helicity
      m_Xpar(  656)=      1.77699d0 !       mass [GeV]   (1.77705 GeV)
      m_Xpar(  657)=          3.5d0 !       WtMax Maximum weight for rejection tau
****** neutrino tau
      m_Xpar(  661)=             16 !       KFlavour
      m_Xpar(  662)=              1 !       NColor
      m_Xpar(  663)=              0 !       3*Q   =3*charge
      m_Xpar(  664)=              1 !       2*T3L =2*Isospin for Left component
      m_Xpar(  665)=              0 !       2*helicity
      m_Xpar(  666)=            0d0 !       mass [GeV]
      m_Xpar(  667)=          3.5d0 !       WtMax Maximum weight for rejection neutrino tau
******************************************************************************
*                     Various switches and options
******************************************************************************
* Some physical switches
*<-i><----data-----><-------------------comments------------------------------>
      m_Xpar( 1001)=            1d0 !  KeyWmp =xpar(1001): choice of W: W- or W+ or both
*                    =0: both W- and W+                   
*                    =1: only W- 
*                    =2: only W+ 
      m_Xpar( 1002)=            1d0 !  KeyQua =xpar(1002): choice of initial quarks
*                    =0: any pair                              
*                    =1: ud                              
*                    =2: cs                              
*                    =3: tb                              
      m_Xpar( 1003)=            1d0 !  KeyLep =xpar(1003): choice of final-state leptons
*                    =0: electrons and muons
*                    =1: electrons
*                    =2: muons
*                    =3: taus
      m_Xpar( 1004)=            0d0 !  KeyWid =xpar(1004): W-boson width scheme
*                    =0: fixed width (RECOMMENDED)
*                    =1: running width 
      m_Xpar( 1005)=            1d0 !  KeyGmu =xpar(1005): input-parameter scheme (IPS)
*                    =0: alpha-scheme
*                    =1: G_mu-scheme (RECOMMENDED)
      m_Xpar( 1006)=            1d0 !  KeyCKM =xpar(1006): quark mixing matrix choice
*                    =0: unit matrix (no mixing)
*                    =1: full CKM matrix (RECOMMENDED)
******************************************************************************
* Radiative corrections switches
*<-i><----data-----><-------------------comments------------------------------>
      m_Xpar( 1011)=             1d0 !  KeyRad =xpar(1011): radiative correction switch
*                    =0: Born level
*                    =1: Above + QED radiative correction in W-decay    
*                    =2: Above + O(alpha) electroweak (EW) corrections 
******************************************************************************
* Some technical switches
*<-i><----data-----><-------------------comments------------------------------>
      m_Xpar( 1021)=            0d0 !  KeyPDF =xpar(1021): choice of PDF parametrization
*                    =0:    parton-level process (no PDFs) 
*                    =1:    some simple, unrealistic parametrization
*                    >1000: PDFs from the library PDFLIB:
*                           KeyPDF = 1000*NGROUP + NSET
*                           See PDFLIB manuals for details. 
      m_Xpar( 1022)=            0d0 !  KeyCol =xpar(1022): choice of collider type      
*                    =0: proton-proton (e.g. LHC)       
*                    =1: proton-antiproton (e.g. TEVATRON)
      m_Xpar( 1023)=            1d0 !  KeyWgt =xpar(1023): event-type choice
*                    =0/1: unweighted/weighted events
      m_Xpar( 1024)=            1d0 !  KeyRnd =xpar(1024)
*                    =1 for RANMAR random number generator (default)
*                    =2 for ECURAN random number generator
*                    =3 for CARRAN random number generator
*<-i><----data-----><-------------------comments------------------------------>
      m_Xpar( 1051)=           11d0 !  IdGenr =xpar(1051): MC generator idetifier (ID)
*                    It is used in weight monitoring and histogramming
*                    in the program. 
*<-i><----data-----><-------------------comments------------------------------>
      m_Xpar( 1061)=            1d0 ! LevPrn =xpar(1061): Print-out level            
*                    =0: Minimum print-out              
*                    =1: Standard print-out (RECOMMENDED)
*                    =2: Verbose print-out, with diagnostics, controls, etc.
******************************************************************************
*EndX
*//////////////////////////////////////////////////////////////////////////////
      END

      SUBROUTINE  WINDEC_SetXpar(i,value)
*     ***********************************
      IMPLICIT NONE
      INCLUDE 'windec.h'
      DOUBLE PRECISION  value
      INTEGER i

      CALL WINDEC_Preset
      m_Xpar(i) = value
      END


      SUBROUTINE  WINDEC_Initialize
*     *********************************************
      IMPLICIT NONE
      INCLUDE 'windec.h'
*//////////////////////////////////////////////////////////////////////////////
*//                                                                          //
*//   Compulsory initialization of WINHAC Mote Carlo Generator               //
*//   INPUT: Ninp,Nout - Input/Output files unit numbers                     //
*//          Xpar - vector of input parameters                               //
*//////////////////////////////////////////////////////////////////////////////
      INTEGER           Ivec1(4),Ivec2(6)    
      DOUBLE PRECISION  Xvec3(4),Dum
      INTEGER           KeyWmp
      INTEGER           Iqup,i

      CALL WINDEC_Preset
      CALL WH_SetIO(m_Ninp,m_Nout)
      m_Nevgen =0
*// Print-out Logo
      CALL WH_Logo
*/////////////////////////////////////////////////////////////////
*//  Initialization of steering parameters, COMMON blocks, etc. //
*/////////////////////////////////////////////////////////////////
      CALL WH_SetPar(m_Xpar)
*/////////////////////////////////////////////////////////////////
*//          Compulsory Initialization of GlibK                 //
*/////////////////////////////////////////////////////////////////
      CALL glimit(50000)
      CALL goutpu(m_Nout)
*/////////////////////////////////////////////////////////////////
      CALL WH_GetKeyTec(Ivec1)
      m_KeyWgt = Ivec1(3)
      CALL WH_GetEmpsw(Xvec3)
      m_Ecm    = Xvec3(1)
      m_WtMax  = Xvec3(4)
*//   Set beams 4-momenta
      CALL WH_SetBeams
*//==============================================
***      CALL WINDEC_MakeCrud(m_XsCrud)
      m_XsCrud = 1d0
      m_ErrCru = 0d0
      CALL WH_SetXsCrud(m_XsCrud,m_ErrCru)
*//==============================================
*/ Initialize generator for parton-level process: qq -> W -> lv 
      CALL WH_GenqqW(-1,m_Ecm,Dum)
*/ Generator's ID
      CALL WH_GetIdGen(m_IdGen)
*/ Important monitor which remembers overall normalization 
*/ and total number of generated events (needed for hist. normalization)
      CALL Gmonit(-1, m_IdGen,1d0,1d0,1d0)
*/ Weights montoring
      DO i = 1,5
         CALL Gmonit(-1,m_IdGen + 10 + i,1d0,1d0,1d0)
      ENDDO
      END


      SUBROUTINE WINDEC_Make(Idqu,Idaq,pqu,paq,
     &                       Idle,Idnu,qle,qnu,Ngam,qgam)
*     ***************************************************
      IMPLICIT NONE
      INCLUDE 'windec.h'
*//////////////////////////////////////////////////////////////////
*//                                                              //
*//   Generation of a single Monte Carlo event.                  //
*//                                                              //
*// Input:  Idqu,Idaq - quark and anti-quark PDG codes           //
*//         pqu,paq   - quark and anti-quark 4-momenta           // 
*// Output: Idle,Idnu - PDG codes of charged lepton and neutrino //
*//         ple,pnu   - 4-momenta of charged lepton and neutrino //
*//         Ngam      - number of radiative photons              //      
*//         qgam      - 4-momenta of radiative photons           //      
*//                                                              //
*//////////////////////////////////////////////////////////////////
      INTEGER          Idqu,Idaq,Idle,Idnu,Ngam
      DOUBLE PRECISION pqu(4),paq(4),qle(4),qnu(4),qgam(100,4)
      DOUBLE PRECISION DrVec(1)
      DOUBLE PRECISION Wtqxx,WtqqW,WtEv,WtOvr,WtNeg,Rn

      m_Nevgen = m_Nevgen+1
*// Set quarks PDG codes and 4-momenta
      CALL WH_SetMomqqW(Idqu,Idaq,pqu,paq)
****************************************************************
*// Return loop for event rejection (unweighted events)
 100  CONTINUE      
      Wtqxx = 1d0
*/-- Save x weight in the event record
      CALL WH_SetWtxx(Wtqxx)
      CALL Gmonit(0,m_IdGen + 12,Wtqxx,m_WtMax,0d0)
*/ Generation of parton-level process: qq -> W -> lv 
      CALL WH_GenqqW( 0, m_Ecm,WtqqW)
      CALL Gmonit(0,m_IdGen + 13,WtqqW,m_WtMax,0d0)
*/ Total best event weight
      WtEv = Wtqxx*WtqqW
*/-- Save the main weight in event record
      CALL WH_SetWtMain(WtEv)
*/ Monitoring weights WtEv > WtMax and WtEv < 0 
      WtOvr = MAX(0d0,WtEv - m_WtMax) 
      CALL Gmonit(0,m_IdGen + 14,WtOvr,0d0,0d0)
      WtNeg = MIN(WtEv,0d0)
      CALL Gmonit(0,m_IdGen + 15,WtNeg,0d0,0d0)  
      IF(m_KeyWgt.EQ.0) THEN    
*/---------------------------------------------------------------------
*/ UNWEIGHTED events, i.e. with WtEv = 1 
*/---------------------------------------------------------------------
         CALL Gmonit(0,m_IdGen,m_XsCrud,m_WtMax,1d0)
         CALL VarRan(DrVec,1)
         Rn = DrVec(1)
         CALL Gmonit(0,m_IdGen + 11,WtEv,m_WtMax,Rn)
*/-- Rejection according to principal weight  
         IF(Rn .GT. WtEv/m_WtMax) GOTO 100       
         WtEv = 1d0    
      ELSE       
*/---------------------------------------------------------------------
*/---------------------------------------------------------------------
*/-- WEIGHTED events      
         CALL Gmonit(0,m_IdGen,m_XsCrud, m_WtMax,1d0)
         CALL Gmonit(0,m_IdGen + 11,WtEv,m_WtMax,0d0)      
      ENDIF 
*/ Lorentz transformations of all four-momenta from W rest-frame 
*/ to LAB frame, i.e. CMS of incoming hadrons
      IF (WtEv .NE. 0d0) CALL WH_LorTraLAB
      END

      SUBROUTINE WINDEC_Finalize
*     **************************
      IMPLICIT NONE
      INCLUDE 'windec.h'
*//////////////////////////////////////////////////////////////////
*//                                                              //
*//   Final printouts and calculation of the total Xsection      //
*//                                                              //
*//////////////////////////////////////////////////////////////////
      DOUBLE PRECISION  Xvec(4),WtMax,Wdum
      DOUBLE PRECISION  Ecm,XsCrud,ErrCru,Dum,AvXcru,ErAvXcr,XNtot 
      INTEGER           Ninp,Nout,IdGen,NevGen,LevPrn
      DOUBLE PRECISION  AvMain,ErAvMain,XseTot,ErXtot,ErMain,ErQxx,AvqqW,ErAvqqW,ErqqW
      INTEGER           Ivec(4),KeyWgt
      DOUBLE PRECISION  XNacc,XNneg,XNove,ErRel
      DOUBLE PRECISION  XevT, Xacc, Xneg, Xove, XevA, XevN, XevO
      DOUBLE PRECISION  AvOve,ErOve,Dum3,XsOve,ErrXO,AvNeg,ErNeg,XsNeg,ErrXN,AvQxx,ErAvQxx
* BoX-FORMATs for nice and flexible outputs
      CHARACTER*80 bxope,bxclo,bxtxt,bxl1i,bxl1l,bxl1f,bxl2f,bxl1g,bxl2g

      CALL WH_GetIO(Ninp,Nout)
      CALL WH_GetIdGen(IdGen)
*/ CMS energy and max. weight
      CALL WH_GetEmpsw(Xvec)
      Ecm   = Xvec(1)
      WtMax = Xvec(4)
*/ Finalize generation of parton-level process: qq -> W -> lv 
      CALL WH_GenqqW(1,Ecm,Wdum)
*/ The "crude" cross section (normalization) and total number of events
      CALL Gmonit(1,IdGen,AvXcru,ErAvXcr,XNtot)
      XsCrud = AvXcru
      ErrCru = 0d0
      NevGen = XNtot
*/ The main MC weight average and its relative statistical error
      CALL Gmonit(1,IdGen + 11,AvMain,ErAvMain,XNtot)    
      ErMain = AvMain*ErAvMain
*/ Total cross section and its  statistical error
      XseTot = XsCrud*AvMain
      ErXtot = XsCrud*ErMain  
*/-- Normalization x-section for UnWeighted events
      CALL WH_GetKeyTec(Ivec)
      KeyWgt = Ivec(3)
      IF (KeyWgt.EQ.0) THEN
         XsCrud = XseTot
         ErrCru = ErXtot
      ENDIF
*/  Store number of events and x-sections in main record wh_common.h
      CALL WH_SetFinXN(NevGen,XseTot,ErXtot,XsCrud,ErrCru)
*/ Number of: generated, accepted, negative-weight and overweighted events 
      XevT = XNtot
      CALL Gmonit(2,IdGen + 11,XNacc,XNneg,XNove) 
      XevA = XNacc
      XevN = XNneg
      XevO = XNove
*/ X-sections for WtEv > WtMax and WtEv < 0 
      CALL Gmonit(1,IdGen + 14,AvOve,ErOve,Dum3)
      XsOve = AvOve/AvMain
      ErrXO = XsOve*ErOve
      CALL Gmonit(1,IdGen + 15,AvNeg,ErNeg,Dum3)
      XsNeg = AvNeg/AvMain
      ErrXN = XsNeg*ErNeg
*/ Component weights: 
*/-- Wtqxx from quarks x1 and x2 generation
      CALL Gmonit(1,IdGen + 12,AvQxx,ErAvQxx,XNtot)  
      ErQxx = AvQxx*ErAvQxx
*/-- WtqqW from parton level process qq -> W -> lv
      CALL Gmonit(1,IdGen + 13,AvqqW,ErAvqqW,XNtot)  
      ErqqW = AvqqW*ErAvqqW
*/ Print-outs if LevPrn > 0  
      CALL WH_GetLevPrn(LevPrn)
      IF (LevPrn .EQ. 0) RETURN
*
      ErRel = ErAvMain
*/ Get box-formats
      CALL WH_GetBxFmts(
     &     bxope,bxclo,bxtxt,bxl1i,bxl1l,bxl1f,bxl2f,bxl1g,bxl2g)
      WRITE(Nout,bxope)         
      WRITE(Nout,bxtxt) '         WINDEC  Final  Report '
      WRITE(Nout,bxtxt) '               Window A        '
      WRITE(Nout,bxtxt) '                               '
      WRITE(Nout,bxl1l) XevT,       'Total no of events','NevTot','A3'
      WRITE(Nout,bxl1l) XevA,       'Accepted    events','NevAcc','A4'
      WRITE(Nout,bxl1l) XevN,       'Wt<0        events','NevNeg','A5'
      WRITE(Nout,bxl2f) XsNeg,ErrXN,'Xsec/Xtot: WT<0   ','XsNeg ','A6'
      WRITE(Nout,bxl1l) XevO,       'Wt > WtMax  events','NevOve','A7'
      WRITE(Nout,bxl2f) XsOve,ErrXO,'Xsec/Xtot:Wt>WtMax','XsOve ','A8'
      WRITE(Nout,bxl1f) WtMax ,     'WtMax             ','WtMax ','A9'
      WRITE(Nout,bxl2f) AvMain,ErMain,'<Wt>           ','AvMain','A10'
      WRITE(Nout,bxl2f) AvMain/WtMax,ErMain/WtMax, 
     &                             'Accept.  Efficiecy','Av/WtM','A11'
      WRITE(Nout,bxclo)         
      
*// Print-outs from  model routine
      CALL WH_Model(1,Wdum)
      END
      
      SUBROUTINE WINDEC_GetWtCrud(WtCrud)
*     **********************************
      IMPLICIT NONE
      INCLUDE 'windec.h'
      DOUBLE PRECISION  WtCrud,Wtxx,WtYFS
*//////////////////////////////////////////////////////////////////
      CALL WH_GetWtCrud(Wtxx,WtYFS)
      WtCrud = Wtxx*WtYFS
      END


      SUBROUTINE WINDEC_MakeCrud(Crud)
*//////////////////////////////////////////////////////////////////
*//   !!!!!!!!!! NOT USED ANYMORE                                //
*//   Calculate normalization xsection                           //
*//                                                              //
*//////////////////////////////////////////////////////////////////
      IMPLICIT NONE
      INCLUDE 'windec.h'
      DOUBLE PRECISION sqq,Crud
*
      DOUBLE PRECISION  pi
      PARAMETER(        pi = 3.1415926535897932d0 )
      INTEGER           KeyWmp,KeyQua,KeyLep,KeyWid,KeyGmu,KeyCKM
      INTEGER           Ivec2(6) 
      DOUBLE PRECISION  Xvec1(4),Xvec2(7),Xmfe(20),Xckm(3,3)
      DOUBLE PRECISION  alpha,G_mu,GeV2nb,aMW,sW2,GaW
      DOUBLE PRECISION  Uij,SigCru,WH_TotWcru,FacNor,alpha_Gmu

      CALL WH_GetKeyPhy(Ivec2)
      KeyWmp = Ivec2(1)         ! W- or W+ or both
      KeyQua = Ivec2(2)         ! quark doublet choice
      KeyLep = Ivec2(3)         ! final-state lepton pair choice  
      KeyWid = Ivec2(4)         ! W width: fixed or running
      KeyGmu = Ivec2(5)         ! IPS: alpha-scheme or G_mu-scheme
      KeyCKM = Ivec2(6)         ! quark mixing-matrix choice
*//==============================================
      sqq = m_Ecm**2
      CALL WH_GetCoCon(Xvec1)
      alpha  = Xvec1(1)
      G_mu   = Xvec1(2)
      GeV2nb = Xvec1(4)
      CALL WH_GetWZsH(Xvec2)
      aMW = Xvec2(1)
      GaW = Xvec2(2)
      sW2 = Xvec2(5)
*/ Overall normalization factor depending on IPS
      IF (KeyGmu .EQ. 0) THEN
*/--- alpha-scheme
         FacNor = GeV2nb
      ELSE
*/--- G_mu-scheme: alpha_Gmu - effective coupling constant at M_W scale
         alpha_Gmu = SQRT(2d0)*G_mu*aMW**2*sW2/pi
         FacNor = GeV2nb * (alpha_Gmu/alpha)**2 
      ENDIF
      IF (KeyLep .EQ. 0) FacNor = 2*FacNor
      CALL WH_GetmfCKM(Xmfe,Xckm)
      Uij = Xckm(1,1)
      Crud = WH_TotWcru(KeyWid,aMW,GaW,sW2,Uij,sqq)
      Crud = Crud*FacNor
      END

